import os
import logging
import warnings
import binascii
import collections
import datetime
import hashlib
import google.auth

from typing import Optional
from urllib.parse import quote, urlencode
from google.oauth2 import service_account
from google.cloud.storage import Client

from terra_notebook_utils import TERRA_DEPLOYMENT_ENV
from terra_notebook_utils.logger import logger


logging.getLogger('google.resumable_media.requests.download').setLevel(logger.getEffectiveLevel())
logging.getLogger('gs_chunked_io.writer').setLevel(logger.getEffectiveLevel())

# Suppress the annoying google gcloud _CLOUD_SDK_CREDENTIALS_WARNING warnings
warnings.filterwarnings('ignore', 'Your application has authenticated using end user credentials')

def get_access_token():
    """Retrieve the access token using the default GCP account returns the same result as
    `gcloud auth print-access-token`.
    """
    if os.environ.get('TERRA_NOTEBOOK_GOOGLE_ACCESS_TOKEN'):
        token = os.environ['TERRA_NOTEBOOK_GOOGLE_ACCESS_TOKEN']
    elif os.environ.get('GOOGLE_APPLICATION_CREDENTIALS'):
        from oauth2client.service_account import ServiceAccountCredentials
        scopes = ['https://www.googleapis.com/auth/userinfo.profile', 'https://www.googleapis.com/auth/userinfo.email']
        creds = ServiceAccountCredentials.from_json_keyfile_name(os.environ['GOOGLE_APPLICATION_CREDENTIALS'],
                                                                 scopes=scopes)
        token = creds.get_access_token().access_token
    else:
        import google.auth.transport.requests
        creds, projects = google.auth.default()
        creds.refresh(google.auth.transport.requests.Request())
        token = creds.token
    return token

def reset_bond_cache():
    from terra_notebook_utils.http import http
    token = get_access_token()
    headers = {
        'authorization': f'Bearer {token}',
        'content-type': 'application/json'
    }
    resp = http.delete(f'http://broad-bond-{TERRA_DEPLOYMENT_ENV}.appspot.com/api/link/v1/fence/', headers=headers)
    print(resp.content)

def get_client(credentials_data: dict | None = None, project: str | None = None):
    kwargs = dict()
    if credentials_data is not None:
        creds = service_account.Credentials.from_service_account_info(credentials_data)
        kwargs['credentials'] = creds
    if project is not None:
        kwargs['project'] = project
    client = Client(**kwargs)
    if credentials_data is None:
        client._credentials.refresh(google.auth.transport.requests.Request())
    return client

def get_signed_url(bucket: str,
                   key: str,
                   sa_credentials: dict | None = None,
                   requester_pays_user_project: Optional[str] = None):
    """
    Returns V4 signed URLs based on a Google bucket, key, and service account credentials.

    If requester_pays_user_project is included, that project will be billed for requester pays objects.

    See: https://cloud.google.com/storage/docs/access-control/signing-urls-manually

    :param str bucket: The name of the Google bucket
    :param str key: The name of the key referencing the Google bucket object
    :param str sa_credentials: A dictionary containing service account credentials with access to the
                               Google key being referenced.
    :param str requester_pays_user_project: (Optional) Name of a Google project to bill if requester pays.
    """
    default_service_account_credentials = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')

    # these are the service account credentials returned from the DRS Resolver
    if sa_credentials:
        creds = service_account.Credentials.from_service_account_info(sa_credentials)
    # if the DRS Resolver did not give us a service account, try the credentials set at GOOGLE_APPLICATION_CREDENTIALS
    elif default_service_account_credentials:
        creds = service_account.Credentials.from_service_account_file(default_service_account_credentials)
    # we can't access the file
    # TODO: implement signed urls for open access files?
    else:
        raise NotImplementedError(
            '\n   Signed URLs are not currently supported for this DRS URI.\n'
            '   If you have a service account that can access this DRS URI, '
            'setting GOOGLE_APPLICATION_CREDENTIALS should enable this.\n'
            '   See: https://cloud.google.com/docs/authentication/production#passing_variable\n'
        )

    canonical_uri = f'/{quote(key.encode("utf-8"), safe=b"/~")}'

    datetime_now = datetime.datetime.utcnow()
    request_timestamp = datetime_now.strftime('%Y%m%dT%H%M%SZ')
    datestamp = datetime_now.strftime('%Y%m%d')

    client_email = creds.service_account_email
    credential_scope = f'{datestamp}/auto/storage/goog4_request'
    credential = f'{client_email}/{credential_scope}'

    host = f'{bucket}.storage.googleapis.com'
    ordered_headers = collections.OrderedDict({'host': host})

    canonical_headers = ''
    for k, v in ordered_headers.items():
        canonical_headers += f'{str(k).lower()}:{str(v).lower()}\n'

    signed_headers = ';'.join([str(k).lower() for k in ordered_headers])

    canonical_query_params = dict()
    canonical_query_params['x-goog-algorithm'] = 'GOOG4-RSA-SHA256'
    canonical_query_params['x-goog-credential'] = credential
    canonical_query_params['x-goog-date'] = request_timestamp
    canonical_query_params['x-goog-expires'] = '3600'
    canonical_query_params['x-goog-signedheaders'] = signed_headers
    if requester_pays_user_project:
        canonical_query_params['userProject'] = requester_pays_user_project

    ordered_query_parameters = collections.OrderedDict(sorted(canonical_query_params.items()))
    canonical_query_string = urlencode(ordered_query_parameters)

    canonical_request = '\n'.join(['GET',
                                   canonical_uri,
                                   canonical_query_string,
                                   canonical_headers,
                                   signed_headers,
                                   'UNSIGNED-PAYLOAD'])
    string_to_sign = '\n'.join(['GOOG4-RSA-SHA256',
                                request_timestamp,
                                credential_scope,
                                hashlib.sha256(canonical_request.encode()).hexdigest()])

    signature = binascii.hexlify(creds.signer.sign(string_to_sign)).decode()
    return f'https://{host}{canonical_uri}?{canonical_query_string}&x-goog-signature={signature}'

def list_bucket(bucket: Optional[str] = None, prefix: str = ''):
    """Lists blobs in "gs://bucket/prefix/"."""
    bucket = bucket or os.environ.get('WORKSPACE_BUCKET')
    if not bucket:
        raise RuntimeError('list_bucket: No bucket specified!  Please provide a bucket name or set WORKSPACE_BUCKET.')
    if bucket.startswith('gs://'):
        bucket = bucket[len('gs://'):]

    # TODO: upgrade Google lib to use new 'gs.get_client().list_blobs(bucket, prefix=prefix)'
    for blob in get_client().bucket(bucket).list_blobs(prefix=prefix):
        yield blob.name



import time
import enum
from math import ceil
from typing import Union

from getm import default_chunk_size
from getm.progress import ProgressBar, ProgressLogger, sizeof_fmt

from terra_notebook_utils.logger import logger


class NotebookProgressBar:
    def __init__(self, name: str, size: int, increments: int=0):
        print(name, flush=True)
        from ipywidgets import HBox, IntProgress, Label
        from IPython.display import display, clear_output
        self.size = size
        self._percent = Label("")
        self._bar = IntProgress(min=0, max=size)
        self._rate = Label("")
        self._duration = Label("")
        ui = HBox([self._percent, self._bar, Label(sizeof_fmt(size)), self._rate, self._duration])
        display(ui)
        logger.debug(f"'{type(self)}' displayed IPython widget '{ui}'.")

    def add(self, sz: int):
        logger.debug(f"'{type(self)}' added progress '{sz}'.")
        duration = time.time() - self._start
        progress = self._bar.value + sz
        self._bar.value = progress
        self._percent.value = "{:3d}%".format((100 * progress) // self.size)
        self._rate.value = f"{sizeof_fmt(progress / duration)}/s"
        self._duration.value = "{:.2f}s".format(duration)

    def __enter__(self):
        logger.debug(f"'{type(self)}' entered context.")
        self._start = time.time()
        self.add(0)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.add(0)
        logger.debug(f"'{type(self)}' exited context.")

class Indicator(enum.Enum):
    log = ProgressLogger
    bar = ProgressBar
    notebook_bar = NotebookProgressBar

    @classmethod
    def get(cls, prog_type: Union[str, "Indicator"], name: str, size: int):
        if isinstance(prog_type, str):
            pt = cls[prog_type]
        elif isinstance(prog_type, cls):
            pt = prog_type
        else:
            raise TypeError(f"unsupported progress indicator type '{type(prog_type)}'")
        return cls._instantiate(pt, name, size)

    @classmethod
    def _instantiate(cls, tp: "Indicator", name: str, size: int):
        if tp in (cls.bar, cls.notebook_bar):
            increments = 20
        elif tp == cls.log:
            increments = ceil(size / default_chunk_size / 2)
        logger.debug(f"Using progress indicator classs '{tp.name}'")
        return tp.value(name, size, increments)



"""
Support for auth with Terra backend services.
"""
from azure.core.exceptions import ClientAuthenticationError
from google.auth.exceptions import DefaultCredentialsError

from terra_notebook_utils import azure_auth, gs, ExecutionPlatform
from terra_notebook_utils.logger import logger
from terra_notebook_utils.utils import get_execution_context


class AuthenticationError(Exception):
    pass


class TerraAuthTokenProvider:
    """
    Provides auth bearer tokens suitable for use with Terra backend services.
    """
    def __init__(self):
        self.execution_context = get_execution_context()

    @staticmethod
    def _identify_valid_access_token() -> str:
        """
        Try to obtain an auth bearer token suitable for use with Terra backend services
        from the Terra supported auth providers. First try Google, then try Azure.
        Return the first successfully obtained token, otherwise raise AuthenticationError.

        :return: auth bearer token suitable for use with Terra backend services
        :raises: AuthenticationError
        """
        try:
            logger.debug("Attempting to obtain a Google access token to use with Terra backend services.")
            google_token = gs.get_access_token()
            logger.debug("Using Google access token to use with Terra backend services.")
            return google_token
        except DefaultCredentialsError as ex:
            logger.debug("Failed to obtain a Google access token to use with Terra backend services.", exc_info=ex)

        try:
            logger.debug("Attempting to obtain a Azure access token to use with Terra backend services.")
            azure_token = azure_auth.get_azure_access_token()
            logger.debug("Using Azure access token to use with Terra backend services.")
            return azure_token
        except ClientAuthenticationError as ex:
            logger.debug("Failed to obtain a Azure access token to use with Terra backend services.", exc_info=ex)

        raise AuthenticationError("Failed to obtain a Google or Azure token to auth with Terra backend services.")

    def get_terra_access_token(self) -> str:
        if self.execution_context.execution_platform == ExecutionPlatform.GOOGLE:
            logger.debug("Using Google default credentials to auth with Terra services.")
            return gs.get_access_token()
        elif self.execution_context.execution_platform == ExecutionPlatform.AZURE:
            logger.debug("Using Azure default credentials to auth with Terra services.")
            return azure_auth.get_azure_access_token()
        else:
            return self._identify_valid_access_token()


# Single instance of TerraAuthTokenProvider.
TERRA_AUTH_TOKEN_PROVIDER = TerraAuthTokenProvider()


def get_terra_access_token() -> str:
    """ Return an auth bearer token suitable for use with Terra backend services.
    :raises: AuthenticationError
    """
    return TERRA_AUTH_TOKEN_PROVIDER.get_terra_access_token()



import io
from functools import wraps
from typing import Generator, Optional, Tuple

from requests.exceptions import HTTPError, ConnectionError
from getm.reader import URLRawReader, URLReaderKeepAlive
from getm.http import http
from getm import checksum, default_chunk_size_keep_alive
from getm.utils import indirect_open

from terra_notebook_utils import blobstore


def catch_blob_not_found(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except (HTTPError, ConnectionError) as ex:
            raise blobstore.BlobNotFoundError(f"Could not find {self.url}") from ex
    return wrapper

def catch_blob_not_found_generator(generator):
    @wraps(generator)
    def wrapper(self, *args, **kwargs):
        try:
            for item in generator(self, *args, **kwargs):
                yield item
        except (HTTPError, ConnectionError) as ex:
            raise blobstore.BlobNotFoundError(f"Could not find {self.url}") from ex
    return wrapper

class URLBlobStore(blobstore.BlobStore):
    schema: Tuple[str, str] = ("http://", "https://")  # type: ignore
    chunk_size = default_chunk_size_keep_alive

    def blob(self, url: str) -> "URLBlob":
        return URLBlob(url)

class URLBlob(blobstore.Blob):
    def __init__(self, url: str, md5: str=""):
        assert url.startswith(URLBlobStore.schema)
        self.url = self.key = url
        self.md5 = md5

    # The next two methods customize pickling behavior. Some modules such as multiprocessing/ProcessPoolExecutor
    # require picklable objects. Let's make pickling snappy.
    # Pickle docs: https://docs.python.org/3/library/pickle.html#object.__getstate
    # multiprocessing docs: https://docs.python.org/3/library/multiprocessing.html#programming-guidelines
    def __getstate__(self):
        return dict(key=self.key, url=self.url)

    def __setstate__(self, state):
        self.__dict__.update(state)

    @catch_blob_not_found
    def get(self) -> bytes:
        with URLRawReader(self.url) as fh:
            return fh.read()

    @catch_blob_not_found
    def open(self, chunk_size: Optional[int]=None) -> io.FileIO:
        return URLRawReader(self.url)

    @catch_blob_not_found_generator
    def download_iter(self, path: str) -> Generator[int, None, None]:
        if self.md5:
            cs = checksum.GETMChecksum(self.md5, "md5")
        else:
            checksums = http.checksums(self.url)
            if 'gs_crc32c' in checksums:
                cs = checksum.GETMChecksum(checksums['gs_crc32c'], "gs_crc32c")
            elif 'gs_md5' in checksum:
                cs = checksum.GETMChecksum(checksums['gs_md5'], "md5")
            elif 'etag' in checksum:
                cs = checksum.GETMChecksum(checksums['gs_md5'], "md5")
            else:
                # TODO: warn user that data integrity checking is disabled
                cs = checksum.GETMChecksum("", checksum.Algorithms.null)

        with indirect_open(path) as fh:
            if http.size(self.url) < URLBlobStore.chunk_size:
                data = self.get()
                cs.update(data)
                fh.write(data)
                yield self.size()
            else:
                for part in self.iter_content():
                    cs.update(part)
                    fh.write(part)
                    yield len(part)
            assert cs.matches(), "Checksum failed!"

    def download(self, path: str):
        for _ in self.download_iter(path):
            pass

    @catch_blob_not_found
    def size(self) -> int:
        return http.size(self.url)

    def iter_content(self) -> blobstore.PartIterator:
        self.size()  # raise BlobNotFoundError
        return URLPartIterator(self.url, URLBlobStore.chunk_size)

class URLPartIterator(blobstore.PartIterator):
    def __init__(self, url: str, chunk_size: int):
        self.url = url
        self.chunk_size = chunk_size

    def __iter__(self):
        if 0 == http.size(self.url):
            yield b""
        else:
            for data in URLReaderKeepAlive.iter_content(self.url, self.chunk_size):
                yield data



import os
import gzip
import tarfile
from typing import Optional

from getm.progress import ProgressBar

from terra_notebook_utils.blobstore import copy_client


_chunk_size = 1024 * 1024 * 32

# Set this for unit tests only
_extract_single_chunk: bool = False

def extract(src_fh, root: Optional[str]=None):
    """Extract a tar.gz archive into the local filesystem, or a GS bucket if `dst_bucket` is provided."""
    root = root or os.getcwd()
    gzip_reader = gzip.GzipFile(fileobj=src_fh)
    tf = tarfile.TarFile(fileobj=gzip_reader)  # type: ignore
    for tarinfo in tf:
        if tarinfo.isfile():
            blob = copy_client.blob_for_url(f"{root}/{tarinfo.name}")
            with ProgressBar(tarinfo.name, tarinfo.size, 40) as progress:
                src_fh = tf.extractfile(tarinfo)
                with blob.part_writer() as writer:
                    while True:
                        data = src_fh.read(_chunk_size)
                        if data:
                            writer.put_part(data)
                            progress.add(len(data))
                        else:
                            break

                        # Hack for speedy unit tests
                        if _extract_single_chunk:
                            return



import os
from dataclasses import dataclass
from enum import Enum

WORKSPACE_NAME = os.environ.get('WORKSPACE_NAME', None)
WORKSPACE_NAMESPACE = os.environ.get('WORKSPACE_NAMESPACE')  # This env var is set in Terra Cloud Environments
WORKSPACE_GOOGLE_PROJECT = os.environ.get('GOOGLE_PROJECT')  # This env var is set in Terra Cloud Environments
TERRA_DEPLOYMENT_ENV = os.environ.get('TERRA_DEPLOYMENT_ENV', 'prod')
MARTHA_URL_VERSION = os.environ.get('MARTHA_URL_VERSION', 'martha_v3')
DRS_RESOLVER_ENDPOINT = os.environ.get("DRS_RESOLVER_ENDPOINT", None)

if not WORKSPACE_GOOGLE_PROJECT:
    WORKSPACE_GOOGLE_PROJECT = os.environ.get('GCP_PROJECT')  # Useful for running outside of notebook
if not WORKSPACE_GOOGLE_PROJECT:
    WORKSPACE_GOOGLE_PROJECT = os.environ.get('GCLOUD_PROJECT')  # Fallback
# For a list of gcloud project related environment variables, see:
# https://cloud.google.com/functions/docs/env-var#environment_variables_set_automatically
WORKSPACE_BUCKET = os.environ.get('WORKSPACE_BUCKET', None)
_GS_SCHEMA = 'gs://'
if WORKSPACE_BUCKET is not None and WORKSPACE_BUCKET.startswith(_GS_SCHEMA):
    WORKSPACE_BUCKET = WORKSPACE_BUCKET[len(_GS_SCHEMA):]  # Chop off the bucket schema

IO_CONCURRENCY = 3

DEPRECATED_MARTHA_URL = f"https://us-central1-broad-dsde-{TERRA_DEPLOYMENT_ENV}.cloudfunctions.net/{MARTHA_URL_VERSION}"
DRSHUB_URL = f"https://drshub.dsde-{TERRA_DEPLOYMENT_ENV}.broadinstitute.org/{DRS_RESOLVER_ENDPOINT}"

if DRS_RESOLVER_ENDPOINT:
    DRS_RESOLVER_URL = DRSHUB_URL
else:
    DRS_RESOLVER_URL = DEPRECATED_MARTHA_URL

class ExecutionEnvironment(Enum):
    TERRA_WORKSPACE = "TERRA_WORKSPACE",  # Executing in a Terra Workspace (on any supported platform)
    OTHER = "OTHER"  # Executing outside a Terra Workspace (e.g., local system)


class ExecutionPlatform(Enum):
    AZURE = "AZURE",  # Executing in an Azure compute environment
    GOOGLE = "GOOGLE",  # Executing in a Google compute environment
    UNKNOWN = "UNKNOWN"  # Execution platform not identified (e.g., local system)


@dataclass
class ExecutionContext:
    execution_environment: ExecutionEnvironment
    execution_platform: ExecutionPlatform



import io
import base64
import multiprocessing
from math import ceil
from concurrent.futures import ThreadPoolExecutor
from typing import Generator, Optional, Union

from getm import checksum, default_chunk_size
from getm.utils import indirect_open
import gs_chunked_io as gscio
from google.cloud.storage import Blob as GSNativeBlob, Bucket as GSNativeBucket
from google.api_core import exceptions as gcp_exceptions

from terra_notebook_utils import blobstore, IO_CONCURRENCY
from terra_notebook_utils import gs as gcp


class GSBlobStore(blobstore.BlobStore):
    schema = "gs://"
    chunk_size = default_chunk_size

    def __init__(self,
                 bucket_name: str,
                 credentials: Optional[dict]=None,
                 billing_project: Optional[str]=None):
        self.bucket_name = bucket_name
        self.billing_project = billing_project
        self.credentials = credentials

    def blob(self, key: str) -> "GSBlob":
        return GSBlob(self.bucket_name, key, self.credentials, self.billing_project)

def _get_native_bucket(bucket: Union[str, GSNativeBucket],
                       credentials: Optional[dict]=None,
                       billing_project: Optional[str]=None) -> GSNativeBucket:
    if isinstance(bucket, str):
        kwargs = dict()
        if billing_project is not None:
            kwargs['user_project'] = billing_project
        bucket = gcp.get_client(credentials, billing_project).bucket(bucket, **kwargs)
    return bucket

def _get_native_blob(bucket: Union[str, GSNativeBucket],
                     key: str,
                     credentials: Optional[dict]=None,
                     billing_project: Optional[str]=None) -> GSNativeBlob:
    bucket = _get_native_bucket(bucket, credentials, billing_project)
    blob = bucket.get_blob(key)
    if blob is None:
        raise blobstore.BlobNotFoundError(f"Could not find gs://{bucket.name}/{key}")
    return blob

class GSBlob(blobstore.Blob):
    def __init__(self,
                 bucket_name: str,
                 key: str,
                 credentials: Optional[dict]=None,
                 billing_project: Optional[str]=None):
        self.bucket_name = bucket_name
        self.key = key
        self.billing_project = billing_project
        self.credentials = credentials

    # The next two methods customize pickling behavior. Some modules such as multiprocessing/ProcessPoolExecutor
    # require picklable objects. Let's make pickling snappy.
    # Pickle docs: https://docs.python.org/3/library/pickle.html#object.__getstate
    # multiprocessing docs: https://docs.python.org/3/library/multiprocessing.html#programming-guidelines
    def __getstate__(self):
        return dict(bucket_name=self.bucket_name,
                    key=self.key,
                    credentials=self.credentials,
                    billing_project=self.billing_project)

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __str__(self):
        return str(dict(bucket_name=self.bucket_name,
                        key=self.key,
                        credentials=self.credentials,
                        billing_project=self.billing_project))

    @property
    def _gs_bucket(self):
        if not getattr(self, "_bucket", None):
            self._bucket = _get_native_bucket(self.bucket_name, self.credentials, self.billing_project)
        return self._bucket

    def _get_native_blob(self) -> GSNativeBlob:
        return _get_native_blob(self._gs_bucket, self.key, self.credentials, self.billing_project)

    @property
    def url(self) -> str:  # type: ignore # TODO https://github.com/python/mypy/issues/4125
        return f"{GSBlobStore.schema}{self.bucket_name}/{self.key}"

    def get(self) -> bytes:
        return self._get_native_blob().download_as_bytes(checksum=None)

    def open(self, chunk_size: Optional[int]=None) -> io.FileIO:
        chunk_size = chunk_size or GSBlobStore.chunk_size
        return self._get_native_blob().open(chunk_size=chunk_size, mode="rb")

    def put(self, data: bytes):
        blob = self._gs_bucket.blob(self.key)
        blob.upload_from_file(io.BytesIO(data))

    def delete(self):
        self._get_native_blob().delete()

    def copy_from_iter(self, src_blob: "GSBlob") -> Generator[int, None, None]:
        assert isinstance(src_blob, type(self))
        if self.url != src_blob.url:
            if not src_blob._gs_bucket.user_project:
                # TODO: always use rewrite when it supports requester pays buckets
                dst_gs_blob = self._gs_bucket.blob(self.key)
                src_gs_blob = src_blob._gs_bucket.blob(src_blob.key)
                token: Optional[str] = None
                total_bytes_written = 0
                while True:
                    try:
                        token, bytes_written, _ = dst_gs_blob.rewrite(src_gs_blob, token)
                    except gcp_exceptions.NotFound as e:
                        raise blobstore.BlobNotFoundError() from e
                    chunk_size = bytes_written - total_bytes_written
                    total_bytes_written = bytes_written
                    yield chunk_size
                    if token is None:
                        break
            else:
                size = src_blob.size()
                if size <= GSBlobStore.chunk_size:
                    self.put(src_blob.get())
                    yield size
                else:
                    with self.part_writer() as writer:
                        for part in src_blob.iter_content():
                            writer.put_part(part)
                            yield len(part)

    def copy_from(self, src_blob: "GSBlob"):
        for _ in self.copy_from_iter(src_blob):
            pass

    def download_iter(self, path: str) -> Generator[int, None, None]:
        with indirect_open(path) as fh:
            cs = self.Hasher()
            for part in self.iter_content():
                fh.write(part)
                cs.update(part)
                yield len(part)
            assert cs.matches(self.cloud_native_checksum())

    def download(self, path: str):
        for _ in self.download_iter(path):
            pass

    def exists(self) -> bool:
        blob = self._gs_bucket.blob(self.key)
        return blob.exists()

    def size(self) -> int:
        return self._get_native_blob().size

    def cloud_native_checksum(self) -> str:
        return self._get_native_blob().crc32c

    @property
    def md5(self) -> str:
        gs_md5 = self._get_native_blob().md5_hash
        return base64.b64decode(gs_md5).hex()

    @property
    def Hasher(self) -> checksum._Hasher:
        return checksum.GSCRC32C

    def iter_content(self) -> blobstore.PartIterator:
        return GSPartIterator(self.bucket_name,
                              self.key,
                              GSBlobStore.chunk_size,
                              credentials=self.credentials,
                              billing_project=self.billing_project)

    def part_writer(self) -> "blobstore.PartWriter":
        return GSPartWriter(self.bucket_name,
                            self.key,
                            credentials=self.credentials,
                            billing_project=self.billing_project)

class GSPartIterator(blobstore.PartIterator):
    def __init__(self,
                 bucket_name: str,
                 key: str,
                 chunk_size: int,
                 credentials: Optional[dict]=None,
                 billing_project: Optional[str]=None):
        self._blob = _get_native_blob(bucket_name, key, credentials, billing_project)
        self.size = self._blob.size
        self.chunk_size = chunk_size
        self._number_of_parts = ceil(self.size / self.chunk_size) if 0 < self.size else 1

    def __iter__(self) -> Generator[bytes, None, None]:
        with ThreadPoolExecutor(max_workers=multiprocessing.cpu_count()) as e:
            q = gscio.async_collections.AsyncQueue(e)
            for _, data in enumerate(gscio.for_each_chunk(self._blob, self.chunk_size, q)):
                yield data

class GSPartWriter(blobstore.PartWriter):
    def __init__(self,
                 bucket_name: str,
                 key: str,
                 credentials: Optional[dict]=None,
                 billing_project: Optional[str]=None):
        super().__init__()
        kwargs = dict()
        if billing_project is not None:
            kwargs['user_project'] = billing_project
        bucket = gcp.get_client(credentials, billing_project).bucket(bucket_name, **kwargs)
        self._executor = ThreadPoolExecutor(max_workers=IO_CONCURRENCY)
        async_set = gscio.async_collections.AsyncSet(self._executor, IO_CONCURRENCY)
        self._part_uploader = gscio.AsyncPartUploader(key, bucket, async_set)
        self._part_number = 0

    def put_part(self, part: bytes):
        self._part_uploader.put_part(self._part_number, part)
        self._part_number += 1

    def close(self):
        self._part_uploader.close()
        self._executor.shutdown()



"""This is the main CLI entry point."""
import sys

# Only the commands modules imported here are available to the CLI
# These must be imported before dispatch
import terra_notebook_utils.cli.commands.config
import terra_notebook_utils.cli.commands.vcf
import terra_notebook_utils.cli.commands.workspace
import terra_notebook_utils.cli.commands.profile
import terra_notebook_utils.cli.commands.drs
import terra_notebook_utils.cli.commands.table
import terra_notebook_utils.cli.commands.workflows
from terra_notebook_utils.cli import dispatch


def main():
    if 2 == len(sys.argv) and "--version" == sys.argv[1].strip():
        from terra_notebook_utils import version
        print(version.__version__)
    else:
        dispatch(sys.argv[1:])



"""Utilities for working with DRS objects."""
import os
import requests

from functools import lru_cache
from collections import namedtuple
from typing import Dict, List, Tuple, Optional, Union, Iterable
from requests import Response

from terra_notebook_utils import WORKSPACE_BUCKET, WORKSPACE_NAME, DRS_RESOLVER_URL, WORKSPACE_NAMESPACE, \
    WORKSPACE_GOOGLE_PROJECT
from terra_notebook_utils import workspace, gs, tar_gz, TERRA_DEPLOYMENT_ENV, _GS_SCHEMA, ExecutionPlatform
from terra_notebook_utils.utils import is_notebook, get_execution_context
from terra_notebook_utils.http import http
from terra_notebook_utils.blobstore.gs import GSBlob
from terra_notebook_utils.blobstore.local import LocalBlob
from terra_notebook_utils.blobstore.url import URLBlob
from terra_notebook_utils.blobstore.progress import Indicator
from terra_notebook_utils.blobstore import Blob, copy_client, BlobNotFoundError
from terra_notebook_utils.logger import logger
from terra_notebook_utils.terra_auth import get_terra_access_token


DRSInfo = namedtuple("DRSInfo", "credentials access_url bucket_name key name size updated checksums")

class DRSResolutionError(Exception):
    pass

class RequesterPaysNotSupported(Exception):
    pass

def _parse_gs_url(gs_url: str) -> Tuple[str, str]:
    if gs_url.startswith(_GS_SCHEMA):
        bucket_name, object_key = gs_url[len(_GS_SCHEMA):].split("/", 1)
        return bucket_name, object_key
    else:
        raise RuntimeError(f'Invalid gs url schema.  {gs_url} does not start with {_GS_SCHEMA}')


def is_requester_pays(drs_urls: Iterable[str]) -> bool:
    """
    Identify if any of the given DRS URIs require Google requester pays access

    :raises: RequesterPaysNotSupported
    """
    for drs_url in drs_urls:
        # Currently (1/2023), Gen3-hosted AnVIL data in GCS is the only
        # DRS data requiring requester pays access.
        # Even this will end in when the AnVIL data is hosted in TDR.
        # Note: If the Gen3 AnVIL DRS URI format is retained by TDR, this function must be updated.

        # The DRS specification (v1.2) states compact identifiers must be lowercase alphanumerical values:
        # https://ga4gh.github.io/data-repository-service-schemas/preview/release/drs-1.2.0/docs/#section/DRS-URIs
        # Yet for historical reasons the DRS URIs minted by Gen3 use uppercase.
        # Perform a case-insensitive comparison.
        anvil_drs_uri_prefix = "drs://dg.ANV0".lower()
        if drs_url.strip().lower().startswith(anvil_drs_uri_prefix):
            if get_execution_context().execution_platform == ExecutionPlatform.AZURE:
                raise RequesterPaysNotSupported(
                    f"Requester pays data access is not supported on the Azure platform. Cannot access: {drs_url}"
                )
            else:
                return True
    return False


@lru_cache()
def enable_requester_pays(workspace_name: Optional[str]=WORKSPACE_NAME,
                          workspace_namespace: Optional[str]=WORKSPACE_NAMESPACE):

    assert get_execution_context().execution_platform != ExecutionPlatform.AZURE, \
        "Requester pays data access is not supported on the Terra Azure platform."

    if not workspace_name:
        raise RuntimeError('Workspace name is not set.  Please set the environment variable '
                           'WORKSPACE_NAME with the name of a valid Terra Workspace.')
    if not workspace_namespace:
        raise RuntimeError("Workspace namespace is not set.  Please set the environment variable "
                           "WORKSPACE_NAMESPACE with the name of your valid Terra Workspace Google Billing Project.")
    import urllib.parse
    encoded_workspace = urllib.parse.quote(workspace_name)
    rawls_url = (f"https://rawls.dsde-{TERRA_DEPLOYMENT_ENV}.broadinstitute.org/api/workspaces/"
                 f"{workspace_namespace}/{encoded_workspace}/enableRequesterPaysForLinkedServiceAccounts")
    logger.info("Enabling requester pays for your workspace. This will only take a few seconds...")
    access_token = get_terra_access_token()

    headers = {
        'authorization': f"Bearer {access_token}",
        'content-type': "application/json"
    }
    resp = http.put(rawls_url, headers=headers)

    if resp.status_code != 204:
        logger.warning(f"Failed to init requester pays for workspace {workspace_namespace}/{workspace_name}: "
                       f"Expected '204', got '{resp.status_code}' for '{rawls_url}'. "
                       "You will not be able to access DRS URIs that interact with requester pays buckets.")

def get_drs(drs_url: str, fields: List[str]) -> Response:
    """Request DRS information from DRS Resolver."""
    access_token = get_terra_access_token()

    headers = {
        'authorization': f"Bearer {access_token}",
        'content-type': "application/json"
    }

    logger.debug(f"Resolving DRS uri '{drs_url}' through '{DRS_RESOLVER_URL}'.")

    json_body = dict(url=drs_url, fields=fields)
    resp = http.post(DRS_RESOLVER_URL, headers=headers, json=json_body)

    if 200 != resp.status_code:
        logger.warning(resp.content)
        error_details = resp.json().get('response', {}).get('text', '')
        raise DRSResolutionError(f"Unexpected response while resolving DRS URI. Expected status 200, got "
                                 f"{resp.status_code}. {error_details}")

    return resp

def info(drs_url: str) -> dict:
    """Return a curated subset of data from `get_drs`."""
    info = get_drs_info(drs_url)
    return dict(name=info.name, size=info.size, updated=info.updated, checksums=info.checksums)

def access(drs_url: str,
           workspace_name: Optional[str]=WORKSPACE_NAME,
           workspace_namespace: Optional[str]=WORKSPACE_NAMESPACE,
           billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> str:
    """Return a signed url for a drs:// URI, if available."""

    if is_requester_pays([drs_url]):
        # We enable requester pays by specifying the workspace/namespace combo, not
        # with the billing project. Rawls then enables requester pays for the attached
        # project, but this won't work if a user specifies a project unattached to
        # the Terra workspace.
        enable_requester_pays(workspace_name, workspace_namespace)

    info = get_drs_info(drs_url, access_url=True)

    if info.access_url:
        drs_resolver_url = info.access_url.strip()
        response = requests.get(drs_resolver_url, headers={'Range': 'bytes=0-1'})
        if response.status_code < 300:
            return drs_resolver_url
        logger.warning('Got an invalid/inaccessible signed URL... attempting to generate an alternate signed URL...')

    url = gs.get_signed_url(bucket=info.bucket_name,
                            key=info.key,
                            sa_credentials=info.credentials)
    # attempt to get the first byte; we'll get an HTTPError error if we need requester pays
    # TODO: hopefully the DRS Resolver will return this information eventually and
    #  we can avoid this check, but even in the meantime, there's probably a better way of doing this
    response = requests.get(url, headers={'Range': 'bytes=0-1'})
    if b'Bucket is a requester pays bucket' in response.content and response.status_code >= 400:
        url = gs.get_signed_url(bucket=info.bucket_name,
                                key=info.key,
                                sa_credentials=info.credentials,
                                requester_pays_user_project=billing_project)
        # one last sanity check
        response = requests.get(url, headers={'Range': 'bytes=0-1'})

    if response.status_code >= 400:
        raise DRSResolutionError(f'Could not generate a valid signed URL for {drs_url} .  Make sure you have access '
                                 f'to this file.  If you believe you have access, relinking your account projects in '
                                 f'your Terra profile may solve this issue (even if it shows your linking is currently '
                                 f'valid).  Otherwise, if you believe you have access and have relinked, please '
                                 f'contact Terra support.')

    return url.strip()

def _get_drs_gs_creds(response: dict) -> Optional[dict]:
    service_account_info = response.get('googleServiceAccount')
    if service_account_info is not None:
        return service_account_info['data']
    else:
        return None

def _drs_info_from_martha_v2(drs_url: str, drs_data: dict) -> DRSInfo:
    """Convert response from martha_v2 to DRSInfo."""
    if 'data_object' in drs_data['dos']:
        data_object = drs_data['dos']['data_object']

        if 'urls' not in data_object:
            raise DRSResolutionError(f"No GS url found for DRS uri '{drs_url}'")
        else:
            data_url = None
            for url_info in data_object['urls']:
                if 'url' in url_info and url_info['url'].startswith(_GS_SCHEMA):
                    data_url = url_info['url']
                    break
            if data_url is None:
                raise DRSResolutionError(f"No GS url found for DRS uri '{drs_url}'")

        bucket_name, key = _parse_gs_url(data_url)
        return DRSInfo(credentials=_get_drs_gs_creds(drs_data),
                       access_url=None,
                       checksums=None,
                       bucket_name=bucket_name,
                       key=key,
                       name=data_object.get('name'),
                       size=data_object.get('size'),
                       updated=data_object.get('updated'))
    else:
        raise DRSResolutionError(f"No metadata was returned for DRS uri '{drs_url}'")

def _drs_info_from_martha_v3_or_drshub(drs_url: str, drs_data: dict) -> DRSInfo:
    """Convert DRS response to DRSInfo."""
    access_url: Optional[str] = None
    if drs_data.get('accessUrl') is not None:
        access_url = drs_data['accessUrl'].get('url')

    if 'gsUri' not in drs_data and not access_url:
        raise DRSResolutionError(f"Neither GS URL or access URL found in DRS response for '{drs_url}'")

    return DRSInfo(credentials=_get_drs_gs_creds(drs_data),
                   access_url=access_url,
                   checksums=drs_data.get('hashes'),
                   bucket_name=drs_data.get('bucket'),
                   key=drs_data.get('name'),
                   name=drs_data.get('fileName'),
                   size=drs_data.get('size'),
                   updated=drs_data.get('timeUpdated'))

def get_drs_info(drs_url: str, access_url: bool = False) -> DRSInfo:
    """Attempt to resolve gs:// url and credentials for a DRS object."""
    assert drs_url.startswith("drs://"), f"Expected DRS URI of the form 'drs://...', got '{drs_url}'"
    fields = ["fileName",
              "hashes",
              "size",
              "gsUri",
              "bucket",
              "name",
              "timeUpdated",
              "googleServiceAccount"]
    # only include this field if necessary because it significantly increases the time for the DRS Resolver to respond
    if access_url:
        fields.append("accessUrl")
    drs_data = get_drs(drs_url, fields).json()
    if 'dos' in drs_data:
        return _drs_info_from_martha_v2(drs_url, drs_data)
    else:
        return _drs_info_from_martha_v3_or_drshub(drs_url, drs_data)

def get_drs_blob(drs_url_or_info: Union[str, DRSInfo],
                 billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> Union[GSBlob, URLBlob]:
    if isinstance(drs_url_or_info, str):
        info = get_drs_info(drs_url_or_info, access_url=True)
    elif isinstance(drs_url_or_info, DRSInfo):
        info = drs_url_or_info
    else:
        raise TypeError(f'Unexpected DRS input type ({type(drs_url_or_info)}): {drs_url_or_info}')
    blob: Union[URLBlob, GSBlob]
    if info.access_url is not None:
        blob = URLBlob(info.access_url, info.checksums.get('md5'))
    else:
        if not (info.credentials or info.bucket_name or info.key):
            raise ValueError(f'DRS information is missing.  Check:\n{info}')
        blob = GSBlob(info.bucket_name, info.key, info.credentials, billing_project)
    return blob

def blob_for_url(url: str, billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> Blob:
    if url.startswith("drs://"):
        return get_drs_blob(url, billing_project)
    else:
        return copy_client.blob_for_url(url)

def head(drs_url: str,
         num_bytes: int = 1,
         workspace_name: Optional[str]=WORKSPACE_NAME,
         workspace_namespace: Optional[str]=WORKSPACE_NAMESPACE,
         billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    """Head a DRS object by byte."""

    if is_requester_pays([drs_url]):
        enable_requester_pays(workspace_name, workspace_namespace)
    try:
        blob = get_drs_blob(drs_url, billing_project)
        with blob.open(chunk_size=num_bytes) as fh:
            the_bytes = fh.read(num_bytes)
    except (DRSResolutionError, BlobNotFoundError) as e:
        raise BlobNotFoundError(f"The DRS URI: '{drs_url}' could not be accessed.") from e
    return the_bytes

def _resolve_bucket_target(url: str, info: DRSInfo) -> Tuple[str, str]:
    bucket_name, pfx = _bucket_name_and_key(url)
    if not pfx or pfx.endswith("/"):
        if pfx.endswith("/"):
            pfx = pfx[:-1]
        basename = info.name or info.key.rsplit("/", 1)[-1]
        key = f"{pfx}/{basename}" if pfx else basename
    else:
        key = pfx
    return bucket_name, key

def _resolve_local_target(filepath: str, info: DRSInfo) -> str:
    # Checking for a file also prevents a batch_copy from overwriting
    # the same file over and over again.
    if os.path.isfile(filepath):
        raise FileExistsError(f'Cannot copy {info.name} to {filepath} because file already exists')
    if filepath.endswith(os.path.sep) or os.path.isdir(filepath):
        filename = info.name or info.key.rsplit("/", 1)[-1]
        filepath = os.path.join(os.path.abspath(filepath), filename)
    return filepath

def _do_copy_drs(drs_uri: str,
                 dst: str,
                 multipart_threshold: int,
                 indicator_type: Indicator):
    dst_blob: Union[GSBlob, URLBlob, LocalBlob]
    src_info = get_drs_info(drs_uri, access_url=True)
    src_blob = get_drs_blob(src_info)
    if dst.startswith("gs://"):
        bucket_name, key = _resolve_bucket_target(dst, src_info)
        dst_blob = GSBlob(bucket_name, key)
    else:
        dst_blob = copy_client.blob_for_url(_resolve_local_target(dst, src_info))
    copy_client._do_copy(src_blob, dst_blob, multipart_threshold, indicator_type)

class DRSCopyClient(copy_client.CopyClient):
    workspace: Optional[str] = None
    workspace_namespace: Optional[str] = None

    def copy(self, drs_uri: str, dst: str):  # type: ignore
        self._queue.put(_do_copy_drs,
                        drs_uri,
                        dst,
                        self.multipart_threshold,
                        self.indicator_type)

def copy(drs_uri: str,
         dst: str,
         indicator_type: Indicator=Indicator.notebook_bar if is_notebook() else Indicator.bar,
         workspace_name: Optional[str]=WORKSPACE_NAME,
         workspace_namespace: Optional[str]=WORKSPACE_NAMESPACE):
    """Copy a DRS object to either the local filesystem, or to a Google Storage location if `dst` starts with
    "gs://".
    """
    if is_requester_pays([drs_uri]):
        enable_requester_pays(workspace_name, workspace_namespace)
    with DRSCopyClient(raise_on_error=True, indicator_type=indicator_type) as cc:
        cc.workspace = workspace_name
        cc.workspace_namespace = workspace_namespace
        cc.copy(drs_uri, dst or ".")

def copy_to_bucket(drs_uri: str,
                   dst_key: str="",
                   dst_bucket_name: Optional[str]=None,
                   indicator_type: Indicator=Indicator.notebook_bar if is_notebook() else Indicator.bar,
                   workspace_name: Optional[str]=WORKSPACE_NAME,
                   workspace_namespace: Optional[str]=WORKSPACE_NAMESPACE):
    """Resolve `drs_url` and copy into user-specified bucket `dst_bucket`.  If `dst_bucket` is None, copy into
    workspace bucket.
    """
    dst_bucket_name = dst_bucket_name or WORKSPACE_BUCKET
    dst_url = f"gs://{dst_bucket_name}"
    if dst_key:
        dst_url += f"/{dst_key}"
    copy(drs_uri, dst_url, indicator_type, workspace_name, workspace_namespace)

def copy_batch(drs_urls: Optional[Iterable[str]] = None,
               dst_pfx: Optional[str] = None,
               workspace_name: Optional[str] = WORKSPACE_NAME,
               workspace_namespace: Optional[str] = WORKSPACE_NAMESPACE,
               indicator_type: Indicator = Indicator.notebook_bar if is_notebook() else Indicator.log,
               manifest: Optional[List[Dict[str, str]]] = None):
    if (manifest is None) == (drs_urls is None):
        raise ValueError("Specify either 'manifest' or 'drs_urls' and 'dst_pfx'")
    elif manifest is not None:
        if dst_pfx is not None:
            raise ValueError('dst_pfx not compatible with manifest')
        copy_batch_manifest(manifest, indicator_type, workspace_name, workspace_namespace)
    elif drs_urls is not None:
        if dst_pfx is None:
            raise ValueError('dst_pfx required with drs_urls')
        copy_batch_urls(drs_urls, dst_pfx, indicator_type, workspace_name, workspace_namespace)
    else:
        assert False

def copy_batch_urls(drs_urls: Iterable[str],
                    dst_pfx: str,
                    indicator_type: Indicator = Indicator.notebook_bar if is_notebook() else Indicator.log,
                    workspace_name: Optional[str] = WORKSPACE_NAME,
                    workspace_namespace: Optional[str] = WORKSPACE_NAMESPACE):
    if is_requester_pays(drs_urls):
        enable_requester_pays(workspace_name, workspace_namespace)
    with DRSCopyClient(indicator_type=indicator_type) as cc:
        cc.workspace = workspace_name
        cc.workspace_namespace = workspace_namespace
        for drs_url in drs_urls:
            cc.copy(drs_url, dst_pfx)

manifest_schema = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "drs_uri": {"type": "string"},
            "dst": {"type": "string"},
        },
        "required": ["drs_uri", "dst"],
    },
}

def copy_batch_manifest(manifest: List[Dict[str, str]],
                        indicator_type: Indicator=Indicator.notebook_bar if is_notebook() else Indicator.log,
                        workspace_name: Optional[str]=WORKSPACE_NAME,
                        workspace_namespace: Optional[str]=WORKSPACE_NAMESPACE):
    from jsonschema import validate
    validate(instance=manifest, schema=manifest_schema)

    drs_uri_list = [item['drs_uri'] for item in manifest]
    if is_requester_pays(drs_uri_list):
        enable_requester_pays(workspace_name, workspace_namespace)

    with DRSCopyClient(indicator_type=indicator_type) as cc:
        cc.workspace = workspace_name
        cc.workspace_namespace = workspace_namespace
        for item in manifest:
            cc.copy(item['drs_uri'], item['dst'])

def extract_tar_gz(drs_url: str,
                   dst: Optional[str]=None,
                   workspace_name: Optional[str]=WORKSPACE_NAME,
                   workspace_namespace: Optional[str]=WORKSPACE_NAMESPACE,
                   billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    """Extract a `.tar.gz` archive resolved by a DRS url. 'dst' may be either a local filepath or a 'gs://' url.
    Default extraction is to the bucket for 'workspace'.
    """
    dst = dst or f"gs://{workspace.get_workspace_bucket(workspace_name)}"

    if is_requester_pays([drs_url]):
        enable_requester_pays(workspace_name, workspace_namespace)
    blob = get_drs_blob(drs_url, billing_project)
    with blob.open() as fh:
        tar_gz.extract(fh, dst)

def _bucket_name_and_key(gs_url: str) -> Tuple[str, str]:
    assert gs_url.startswith("gs://")
    parts = gs_url[5:].split("/", 1)
    if 1 >= len(parts) or not parts[1]:
        bucket_name = parts[0]
        key = ""
    else:
        bucket_name, key = parts
    return bucket_name, key



import os
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from typing import Optional, Union

from getm import default_chunk_size
from getm.concurrent.collections import ConcurrentHeap

from terra_notebook_utils.blobstore.gs import GSBlobStore, GSBlob
from terra_notebook_utils.blobstore.url import URLBlobStore, URLBlob
from terra_notebook_utils.blobstore.local import LocalBlobStore, LocalBlob
from terra_notebook_utils.blobstore.progress import Indicator
from terra_notebook_utils.blobstore import BlobstoreChecksumError
from terra_notebook_utils.logger import logger


AnyBlobStore = Union[GSBlobStore, URLBlobStore, LocalBlobStore]
AnyBlob = Union[GSBlob, URLBlob, LocalBlob]
CloudBlob = GSBlob

def _download(src_blob: AnyBlob, dst_blob: LocalBlob, indicator_type: Indicator):
    logger.debug(f"Starting download {src_blob.url} to {dst_blob.url}")
    dirname = os.path.dirname(dst_blob.url)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    # The download methods for each Blob is expected to compute checksums
    with Indicator.get(indicator_type, dst_blob.url, src_blob.size()) as indicator:
        for part_size in src_blob.download_iter(dst_blob.url):
            indicator.add(part_size)

def _copy_intra_cloud(src_blob: AnyBlob, dst_blob: AnyBlob, indicator_type: Indicator):
    # In general it is not necessary to compute checksums for intra cloud copies. The Storage services will do that for
    # us.
    # However, S3Etags depend on the part layout. Either ensure source and destination part layouts are the same, or
    # compute the destination S3Etag on the fly.
    logger.debug(f"Starting intra-cloud {src_blob.url} to {dst_blob.url}")
    assert isinstance(src_blob, type(dst_blob))
    with Indicator.get(indicator_type, dst_blob.url, src_blob.size()) as indicator:
        for part_size in dst_blob.copy_from_iter(src_blob):  # type: ignore
            indicator.add(part_size)
    if src_blob.cloud_native_checksum() != dst_blob.cloud_native_checksum():
        logger.error(f"Checksum failed for {src_blob.url} to {dst_blob.url}")
        raise BlobstoreChecksumError()

def _copy_oneshot_passthrough(src_blob: Union[URLBlob, CloudBlob], dst_blob: CloudBlob, indicator_type: Indicator):
    logger.debug(f"Starting oneshot passthrough {src_blob.url} to {dst_blob.url}")
    with Indicator.get(indicator_type, dst_blob.url, src_blob.size()) as indicator:
        data = src_blob.get()
        dst_blob.put(data)
        indicator.add(len(data))
    if dst_blob.md5 != src_blob.md5:
        logger.error(f"Checksum failed for {src_blob.url} to {dst_blob.url}")
        raise BlobstoreChecksumError()

def _copy_multipart_passthrough(src_blob: Union[URLBlob, CloudBlob], dst_blob: CloudBlob, indicator_type: Indicator):
    logger.debug(f"Starting multipart passthrough {src_blob.url} to {dst_blob.url}")
    with Indicator.get(indicator_type, dst_blob.url, src_blob.size()) as indicator:
        with dst_blob.part_writer() as writer:
            for part in src_blob.iter_content():
                writer.put_part(part)
                indicator.add(len(part))
    if dst_blob.md5 != src_blob.md5:
        logger.error(f"Checksum failed for {src_blob.url} to {dst_blob.url}")
        raise BlobstoreChecksumError()

def _do_copy(src_blob: AnyBlob, dst_blob: AnyBlob, multipart_threshold: int, indicator_type: Indicator):
    try:
        if isinstance(dst_blob, LocalBlob):
            _download(src_blob, dst_blob, indicator_type)
        elif isinstance(src_blob, type(dst_blob)):
            _copy_intra_cloud(src_blob, dst_blob, indicator_type)
        elif isinstance(dst_blob, CloudBlob):
            # The following assert prevents LocalBlob -> CloudBlob, i.e. upload. This should be removed in TNU is
            # expected to upload local files to cloud locations. Checksumming logic will also need updates.
            assert isinstance(src_blob, (URLBlob, CloudBlob))
            if src_blob.size() <= multipart_threshold:
                _copy_oneshot_passthrough(src_blob, dst_blob, indicator_type)
            else:
                _copy_multipart_passthrough(src_blob, dst_blob, indicator_type)
        else:
            raise TypeError(f"Cannot handle copy operation for blob types '{type(src_blob)}' and '{type(dst_blob)}'")
        logger.debug(f"Copied {src_blob.url} to {dst_blob.url}")
    except Exception:
        logger.exception(f"copy failed: '{src_blob.url}' to '{dst_blob.url}'")
        # FIXME: The finally block is useless, but what was the intent? Probably idempotent delete
        try:
            dst_blob.delete()
        finally:
            pass
        raise

def blob_for_url(url: str) -> AnyBlob:
    if url.startswith(GSBlobStore.schema):
        parts = url[len(GSBlobStore.schema):].split("/", 1)
        if 2 != len(parts):
            raise ValueError(f"Incorrect GS url '{url}'")
        bucket, key = parts
        return GSBlob(bucket, key)
    elif url.startswith(URLBlobStore.schema):
        return URLBlob(url)
    else:
        return LocalBlob(os.path.sep, os.path.abspath(url))

class CopyClient:
    multipart_threshold = default_chunk_size

    def __init__(self,
                 concurrency: int=multiprocessing.cpu_count(),
                 raise_on_error: bool=False,
                 indicator_type: Optional[Indicator]=None):
        """If 'raise_on_error' is False, all copy operations will be attempted even if one or more operations error. If
        'raise_on_error' is True, the first error encountered will be raise and all scheduled operations will be
        canceled.
        """
        self._executor = ProcessPoolExecutor(max_workers=concurrency)
        self._queue = ConcurrentHeap(self._executor, concurrency)
        self.raise_on_error = raise_on_error
        self.indicator_type = indicator_type or Indicator.log

    def copy(self, src: Union[str, AnyBlob], dst: Union[str, AnyBlob]):
        if isinstance(src, str):
            src = blob_for_url(src)
        if isinstance(dst, str):
            dst = blob_for_url(dst)
        priority = -src.size()
        self._queue.priority_put(priority, _do_copy, src, dst, self.multipart_threshold, self.indicator_type)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        try:
            for f in self._queue.iter_futures():
                if self.raise_on_error:
                    try:
                        f.result()
                    except Exception:
                        self._queue.abort()
                        raise
        finally:
            self._executor.shutdown()

def copy(src: Union[str, AnyBlob], dst: Union[str, AnyBlob]):
    with CopyClient(indicator_type=Indicator.log) as client:
        client.copy(src, dst)



import json
import argparse

from terra_notebook_utils import vcf, WORKSPACE_GOOGLE_PROJECT
from terra_notebook_utils.cli import dispatch
from terra_notebook_utils.drs import blob_for_url


vcf_cli = dispatch.group("vcf", help=vcf.__doc__, arguments={
    "path": dict(
        help="local path, gs://, or drs://"
    ),
    "--billing-project": dict(
        type=str,
        required=False,
        default=WORKSPACE_GOOGLE_PROJECT,
        help=("The billing project for GS requests. "
              "If omitted, the environment variables GOOGLE_PROJECT, "
              "GCP_PROJECT, AND GCLOUD_PROJECT will be in that order.")
    ),
})

@vcf_cli.command("head")
def head(args: argparse.Namespace):
    """
    Output VCF header.
    """
    blob = blob_for_url(args.path, args.billing_project)
    info = vcf.VCFInfo.with_blob(blob)
    info.print_header()

@vcf_cli.command("samples")
def samples(args: argparse.Namespace):
    """
    Output VCF samples.
    """
    blob = blob_for_url(args.path, args.billing_project)
    info = vcf.VCFInfo.with_blob(blob)
    print(json.dumps(info.samples, indent=2))

@vcf_cli.command("stats")
def stats(args: argparse.Namespace):
    """
    Output VCF stats.
    """
    blob = blob_for_url(args.path, args.billing_project)
    info = vcf.VCFInfo.with_blob(blob)
    stats = {
        'first data line chromosome': info.chrom,
        'length associated with first data line chromosome': info.length,
        'number of samples': len(info.samples),
        'size': blob.size(),
    }
    print(json.dumps(stats, indent=2))



import sys
import json
import argparse
from typing import Any, Dict

from terra_notebook_utils import drs
from terra_notebook_utils.cli import dispatch, CLIConfig
from terra_notebook_utils.drs import DRSResolutionError


drs_cli = dispatch.group("drs", help=drs.__doc__)

workspace_args: Dict[str, Dict[str, Any]] = {
    "--workspace": dict(
        type=str,
        default=None,
        help="workspace name. If not provided, the configured CLI workspace will be used"
    ),
    "--workspace-namespace": dict(
        type=str,
        default=CLIConfig.info['workspace_namespace'],
        help=("The workspace namespace represents the parent containing the workspace "
              "(the Terra billing project) "
              "If omitted, the CLI configured `workspace_namespace` will be used. ")
    )
}

@drs_cli.command("copy", arguments={
    "drs_url": dict(type=str),
    "dst": dict(type=str, help="local file path, or Google Storage location if prefixed with 'gs://'"),
    ** workspace_args,
})
def drs_copy(args: argparse.Namespace):
    """
    Copy drs:// object to local file or Google Storage bucket

    If 'dst' is suffixed with "/", the destination is assumed to be a folder and the file name is
    derived from the drs response and appended to 'dst'. Otherwise the destination is assumed
    to be absolute.

    examples:
        tnu drs copy drs://my-drs-id /tmp/doom  # copy to /tmp/doom
        tnu drs copy drs://my-drs-id /tmp/doom/  # copy to /tmp/doom/{file-name-from-drs-resolution}
        tnu drs copy drs://my-drs-id gs://my-cool-bucket/my-cool-bucket-key
        tnu drs copy drs://my-drs-id gs://my-cool-bucket/my-cool-bucket-key/
    """
    args.workspace, args.workspace_namespace = CLIConfig.resolve(args.workspace, args.workspace_namespace)
    kwargs: Dict[str, Any] = dict(workspace_name=args.workspace, workspace_namespace=args.workspace_namespace)
    if CLIConfig.progress_indicator_type() is not None:
        kwargs['indicator_type'] = CLIConfig.progress_indicator_type()
    drs.copy(args.drs_url, args.dst, **kwargs)

@drs_cli.command("copy-batch", arguments={
    "drs_uris": dict(type=str, nargs="*", help="space separated list of drs:// URIs"),
    "--dst": dict(type=str, default=None, help="local directory, or Google Storage location if prefixed with 'gs://'"),
    "--manifest": dict(type=str, default=None, help="filepath to JSON manifest."),
    ** workspace_args,
})
def drs_copy_batch(args: argparse.Namespace):
    """
    Copy several drs:// objects to local directory or Google Storage bucket
    examples:
        tnu drs copy-batch drs://my-drs-1 drs://my-drs-2 drs://my-drs-3 --dst /tmp/doom/
        tnu drs copy-batch drs://my-drs-1 drs://my-drs-2 drs://my-drs-3 --dst gs://my-cool-bucket/my-cool-folder
        tnu drs copy-batch --manifest manifest.json

    When not using a manifest, 'dst' is treated as a folder, and file names are derived from the drs response.
    Otherwise, in a manifest, 'dst' can either be a folder (if suffixed with "/"), or an absolute path, e.g.
    '/home/me/my-file-name.vcf.gz' or 'gs://bucket-name/pfx/my-file.vcf.gz'.

    example manifest.json:
    [
      {
        "drs_uri": "drs://my/cool/drs/uri",
        "dst": "/path/to/local/dir/"
      },
      {
        "drs_uri": "drs://my/cool/drs/uri",
        "dst": "gs://my-cook-bucket/my-cool-prefix"
      }
    ]
    """
    args.workspace, args.workspace_namespace = CLIConfig.resolve(args.workspace, args.workspace_namespace)
    kwargs: Dict[str, Any] = dict(workspace_name=args.workspace, workspace_namespace=args.workspace_namespace)
    if CLIConfig.progress_indicator_type() is not None:
        kwargs['indicator_type'] = CLIConfig.progress_indicator_type()
    if args.drs_uris:
        assert args.manifest is None, "Cannot use 'drs_uris' with '--manifest'"
        assert args.dst is not None, "Must specify a destination with '--dst'"
        drs.copy_batch_urls(args.drs_uris, args.dst, **kwargs)
    elif args.manifest:
        with open(args.manifest) as fh:
            manifest = json.loads(fh.read())
        drs.copy_batch_manifest(manifest, **kwargs)
    else:
        raise RuntimeError("Must supply either 'drs_uris' or '--manifest'")

@drs_cli.command("head", arguments={
    "drs_url": dict(type=str),
    "--bytes": dict(type=int, required=False, default=1, help="Number of bytes to fetch."),
    ** workspace_args,
})
def drs_head(args: argparse.Namespace):
    """
    Print the first bytes of a drs:// object.

    Example:
        tnu drs head drs://crouching-drs-hidden-access
    """
    args.workspace, args.workspace_namespace = CLIConfig.resolve(args.workspace, args.workspace_namespace)
    the_bytes = drs.head(args.drs_url,
                         num_bytes=args.bytes,
                         workspace_name=args.workspace,
                         workspace_namespace=args.workspace_namespace)
    # sys.stdout.buffer is used outside of a python notebook since that's the standard non-lossy way
    # to write/display bytes; sys.stdout.buffer is not available inside of a python notebook
    # though, as sys.stdout is a ipykernel.iostream.OutStream object:
    # https://github.com/ipython/ipykernel/blob/master/ipykernel/iostream.py#L265
    # so we use bare sys.stdout and rely on the ipykernel method's lossy unicode stream
    stdout_buffer = getattr(sys.stdout, 'buffer', sys.stdout)
    stdout_buffer.write(the_bytes)

@drs_cli.command("extract-tar-gz", arguments={
    "drs_url": dict(type=str),
    "dst": dict(type=str,
                nargs='?',
                default=None,
                help=("Root of extracted archive. This may be a local filepath or a 'gs://' url."
                      "If not provided, default extraction is to the bucket for 'workspace'")),
    ** workspace_args,
})
def drs_extract_tar_gz(args: argparse.Namespace):
    """
    Extract a `tar.gz` archive resolved by DRS into a Google Storage bucket.
    example:
        tnu drs extract-tar-gz drs://my-tar-gz gs://my-dst-bucket/root
    """
    args.workspace, args.workspace_namespace = CLIConfig.resolve(args.workspace, args.workspace_namespace)
    drs.extract_tar_gz(args.drs_url, args.dst, args.workspace, args.workspace_namespace)

@drs_cli.command("info", arguments={
    "drs_url": dict(type=str),
})
def drs_info(args: argparse.Namespace):
    """
    Get information about a drs:// URI
    """
    info = drs.info(args.drs_url)
    print(json.dumps(info, indent=2))

@drs_cli.command("access", arguments={
    "drs_url": dict(type=str),
    ** workspace_args,
})
def drs_access(args: argparse.Namespace):
    """
    Get a signed url for a drs:// URI
    """
    args.workspace, args.workspace_namespace = CLIConfig.resolve(args.workspace, args.workspace_namespace)
    signed_url = drs.access(args.drs_url, args.workspace, args.workspace_namespace)
    print(signed_url.strip())

@drs_cli.command("credentials", arguments={
    "drs_url": dict(type=str),
})
def drs_credentials(args: argparse.Namespace):
    """
    Return the credentials needed to access a DRS url.
    """
    try:
        info = drs.get_drs_info(args.drs_url)
    except DRSResolutionError:
        raise
    except Exception:
        raise

    print(json.dumps(info.credentials, indent=2))



from typing import BinaryIO

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


default_retry = Retry(total=10,
                      status_forcelist=[429, 500, 502, 503, 504],
                      method_whitelist=["HEAD", "GET"])

def http_session(session: requests.Session | None = None, retry: Retry | None = None) -> requests.Session:
    session = session or requests.Session()
    retry = retry or default_retry
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("https://", adapter)
    session.mount("http://", adapter)
    return session

# Instantiate a default session. It's useful to have a common session to take advantage of connection pooling.
# Users can modify stuff by replacing this session or instantiate a new one. Note that sessions can be used
# as context managers:
# with http_session(...) as http:
#    http.get(...)
http = http_session()



import io
from collections import namedtuple
from typing import Any, Generator, Optional

from getm import checksum


MiB = 1024 * 1024

class BlobStore:
    schema: str
    chunk_size: int

    def list(self, prefix: str=""):
        raise NotImplementedError()

    def blob(self, key: str):
        raise NotImplementedError()

class Blob:
    key: str
    url: str

    def get(self) -> bytes:
        raise NotImplementedError()

    def open(self, chunk_size: Optional[int]=None) -> io.FileIO:
        raise NotImplementedError()

    def put(self, data: bytes):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    def copy_from_iter(self, src_blob: Any) -> Generator[int, None, None]:
        """Intra-cloud copy. This yields the size of parts copied, and must be iterated"""
        raise NotImplementedError()

    def copy_from(self, src_blob: Any) -> Generator[int, None, None]:
        """Intra-cloud copy."""
        raise NotImplementedError()

    def download_iter(self, path: str) -> Generator[int, None, None]:
        """This yields the size of parts downloaded, and must be iterated"""
        raise NotImplementedError()

    def download(self, path: str) -> Generator[int, None, None]:
        raise NotImplementedError()

    def exists(self) -> bool:
        raise NotImplementedError()

    def size(self) -> int:
        raise NotImplementedError()

    def cloud_native_checksum(self) -> str:
        raise NotImplementedError()

    def hash_class(self) -> checksum._Hasher:
        return checksum.GSCRC32C

    def iter_content(self) -> "PartIterator":
        raise NotImplementedError()

    def part_writer(self) -> "PartWriter":
        raise NotImplementedError()

class PartIterator:
    def __len__(self):
        raise NotImplementedError()

    def __iter__(self) -> Generator[bytes, None, None]:
        raise NotImplementedError()

class PartWriter:
    def put_part(self, part: bytes):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

class BlobStoreError(Exception):
    pass

class BlobNotFoundError(BlobStoreError):
    pass

class BlobStoreUnknownError(BlobStoreError):
    pass

class BlobstoreChecksumError(BlobStoreError):
    pass



"""
Microsoft Azure identity/auth support.

See:
https://azuresdkdocs.blob.core.windows.net/$web/python/azure-identity/1.12.0/index.html
https://learn.microsoft.com/en-us/python/api/azure-identity/azure.identity.defaultazurecredential?view=azure-python
https://github.com/Azure/azure-sdk-for-python/blob/main/sdk/identity/azure-identity/azure/identity/_credentials/default.py
"""
import logging
import os
from typing import Optional

from azure.identity import DefaultAzureCredential
from terra_notebook_utils.logger import logger


# Single instance of DefaultAzureCredential that initialized lazily.
# The instance is treated as threadsafe and reusable.
# The Azure documentation is silent on thread safety.
# Based on scanning the code, it appears to be threadsafe.
# See: https://github.com/Azure/azure-sdk-for-python/blob/main/sdk/identity/
#              azure-identity/azure/identity/_credentials/default.py
_AZURE_CREDENTIAL: Optional[DefaultAzureCredential] = None


def _set_azure_identity_logging_level(level: int) -> None:
    """ Set the logging level for modules participating the Azure default credential flow """
    import azure.identity
    logging.getLogger(azure.identity._credentials.environment.__name__).setLevel(level)
    logging.getLogger(azure.identity._credentials.managed_identity.__name__).setLevel(level)
    logging.getLogger(azure.identity._credentials.chained.__name__).setLevel(level)
    logging.getLogger("azure.core.pipeline.policies.http_logging_policy").setLevel(level)


# Suppress extraneous azure-identity INFO level logging
_set_azure_identity_logging_level(logging.WARNING)


def _get_default_credential() -> DefaultAzureCredential:
    """
    Instantiate DefaultAzureCredential lazily if/when needed.

    Note: It would not need to be instantiated this way, as
    # no exception is raised even if Azure credentials are not configured.
    :return: Reference to instance of DefaultAzureCredential
    """

    # Should a more sophisticated Singleton pattern be used instead?
    global _AZURE_CREDENTIAL
    if not _AZURE_CREDENTIAL:
        _AZURE_CREDENTIAL = DefaultAzureCredential()
    return _AZURE_CREDENTIAL


def get_azure_access_token() -> str:
    """
    Return an Azure access token.

    raises ClientAuthenticationError
    """
    if os.environ.get('TERRA_NOTEBOOK_AZURE_ACCESS_TOKEN'):
        logger.debug("Using Azure token configured using 'TERRA_NOTEBOOK_AZURE_ACCESS_TOKEN'")
        token = os.environ['TERRA_NOTEBOOK_AZURE_ACCESS_TOKEN']
    else:
        logger.debug("Requesting Azure default credentials token.")
        token_scope = "https://management.azure.com/.default"
        azure_token = _get_default_credential().get_token(token_scope)
        logger.debug("Using Azure default credentials token.")
        token = azure_token.token
    return token



import threading
from http.server import HTTPServer, BaseHTTPRequestHandler
from typing import Tuple


class ThreadedLocalServer(threading.Thread):
    """Run an http server on a separate thread."""
    def __init__(self, handler_class: BaseHTTPRequestHandler, address: Tuple[str, int]=('', 8000)):
        super().__init__(daemon=True)
        self.address = address
        self._handler_class = handler_class
        self._server = None
        self._server_ready = threading.Event()

    def start(self):
        super().start()
        self._server_ready.wait()

    def run(self):
        self._server = HTTPServer(self.address, self._handler_class)
        self._server_ready.set()
        self._server.serve_forever()

    def shutdown(self):
        if self._server is not None:
            self._server.shutdown()
        self.join(timeout=5)
        assert not self.is_alive(), "Failed to join thread"

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.shutdown()



import io
import os
import shutil
from math import ceil
from functools import wraps
from typing import Generator, Optional

from getm import default_chunk_size

from terra_notebook_utils import blobstore


def catch_blob_not_found(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except FileNotFoundError as ex:
            raise blobstore.BlobNotFoundError(f"Could not find {self.url}") from ex
    return wrapper

def catch_blob_not_found_generator(generator):
    @wraps(generator)
    def wrapper(self, *args, **kwargs):
        try:
            for item in generator(self, *args, **kwargs):
                yield item
        except FileNotFoundError as ex:
            raise blobstore.BlobNotFoundError(f"Could not find {self.url}") from ex
    return wrapper

class LocalBlobStore(blobstore.BlobStore):
    chunk_size = default_chunk_size

    def __init__(self, basepath: str):
        self.bucket_name = basepath

    def list(self, prefix: str="") -> Generator["LocalBlob", None, None]:
        root = os.path.join(self.bucket_name, prefix)
        if root.endswith(os.path.sep):
            root = root[:-len(os.path.sep)]
        for (dirpath, dirnames, filenames) in os.walk(root):
            for filename in filenames:
                relpath = os.path.relpath(os.path.join(dirpath, filename), self.bucket_name)
                yield LocalBlob(self.bucket_name, relpath)

    def blob(self, key: str) -> "LocalBlob":
        return LocalBlob(self.bucket_name, key)

class LocalBlob(blobstore.Blob):
    def __init__(self, basepath: str, relpath: str):
        assert basepath == os.path.abspath(basepath)
        self.bucket_name = basepath
        self.key = relpath
        self._path = os.path.join(basepath, relpath)
        self.makedirs = True  # create subdirectories if needed during write operations

    # The next two methods customize pickling behavior. Some modules such as multiprocessing/ProcessPoolExecutor
    # require picklable objects. Let's make pickling snappy.
    # Pickle docs: https://docs.python.org/3/library/pickle.html#object.__getstate
    # multiprocessing docs: https://docs.python.org/3/library/multiprocessing.html#programming-guidelines
    def __getstate__(self):
        return dict(bucket_name=self.bucket_name, key=self.key, _path=self._path)

    def __setstate__(self, state):
        self.__dict__.update(state)

    @property
    def url(self) -> str:  # type: ignore # TODO https://github.com/python/mypy/issues/4125
        return self._path

    @catch_blob_not_found
    def get(self) -> bytes:
        with open(self._path, "rb") as fh:
            return fh.read()

    @catch_blob_not_found
    def open(self, chunk_size: Optional[int]=None) -> io.FileIO:
        return open(self._path, "rb")  # type: ignore  # this is technically BinaryIO but should be compatible

    def put(self, data: bytes):
        if self.makedirs:
            os.makedirs(os.path.dirname(self._path), exist_ok=True)
        with open(self._path, "wb") as fh:
            fh.write(data)

    @catch_blob_not_found
    def delete(self):
        os.remove(self._path)

    @catch_blob_not_found_generator
    def copy_from_iter(self, src_blob: "LocalBlob") -> Generator[int, None, None]:
        """
        Intra-cloud copy
        """
        assert isinstance(src_blob, type(self))
        if self.url != src_blob.url:
            shutil.copyfile(src_blob._path, self._path)
        yield self.size()

    def copy_from(self, src_blob: "LocalBlob"):
        for part in self.copy_from_iter(src_blob):
            pass

    def download_iter(self, target: str) -> Generator[int, None, None]:
        if not os.path.isfile(self._path):
            raise blobstore.BlobNotFoundError(f"Could not find {self.url}")
        if self._path != target:
            shutil.copyfile(self._path, target)
        yield self.size()

    def download(self, target: str):
        for _ in self.download_iter(target):
            pass

    def exists(self) -> bool:
        if os.path.isdir(self._path):
            raise ValueError(f"'{self._path}' exists but is not a file!")
        return os.path.isfile(self._path)

    @catch_blob_not_found
    def size(self) -> int:
        return os.path.getsize(self._path)

    def iter_content(self) -> blobstore.PartIterator:
        return LocalPartIterator(self._path)

    def part_writer(self) -> "LocalPartWriter":
        return LocalPartWriter(self._path, self.makedirs)

class LocalPartIterator(blobstore.PartIterator):
    def __init__(self, path: str):
        try:
            self.size = os.path.getsize(path)
        except FileNotFoundError:
            raise blobstore.BlobNotFoundError(f"Could not find {path}")
        self.chunk_size = LocalBlobStore.chunk_size
        self._number_of_parts = ceil(self.size / self.chunk_size) if 0 < self.size else 1
        self.handle = open(path, "rb")

    def __len__(self):
        return self._number_of_parts

    def __iter__(self) -> Generator[bytes, None, None]:
        for _ in range(self._number_of_parts):
            yield self.handle.read(self.chunk_size)

    def close(self):
        if hasattr(self, "handle"):
            self.handle.close()

    def __del__(self):
        self.close()

class LocalPartWriter(blobstore.PartWriter):
    """This provides a consistent interface for blobs, but is mostly just a wrapper over a normal file handle."""
    def __init__(self, filepath: str, makedirs: bool):
        if makedirs:
            os.makedirs(os.path.dirname(filepath), exist_ok=True)
        self._fh = open(filepath, "wb")

    def put_part(self, part: bytes):
        self._fh.write(part)

    def close(self):
        self._fh.close()



"""
Configure the CLI
"""
import typing
import argparse

from terra_notebook_utils.cli import CLIConfig, dispatch


config_cli = dispatch.group("config", help=__doc__)


@config_cli.command("set-workspace", arguments={
    "workspace": dict()
})
def set_config_workspace(args: argparse.Namespace):
    """
    Set workspace for cli commands
    """
    CLIConfig.info["workspace"] = args.workspace
    CLIConfig.write()

@config_cli.command("set-workspace-namespace", arguments={
    "workspace_namespace": dict(type=str)
})
def set_config_workspace_namespace(args: argparse.Namespace):
    """
    Set workspace namespace for cli commands
    """
    CLIConfig.info["workspace_namespace"] = args.workspace_namespace
    CLIConfig.write()

@config_cli.command("set-copy-progress-indicator-type", arguments={
    "copy_progress_indicator_type": dict(type=str, choices=["auto", "log"])
})
def set_indicator_type(args: argparse.Namespace):
    """
    Set the indicator type for DRS copy operations.

    When 'copy-progress-indicator-type' is set to 'auto', terra-notebook-utils chooses the most appropriate
    indicator type for copy operations.
    """
    CLIConfig.info["copy_progress_indicator_type"] = args.copy_progress_indicator_type
    CLIConfig.write()

@config_cli.command("print")
def config_print(args: argparse.Namespace):
    """
    Print the CLI configuration.
    """
    import json
    print(json.dumps(CLIConfig.info, indent=2))
